;;; eev.el -- add support for e-scripts in Emacs.

;; Copyright (C) 1999,2000,2001,2002,2003,2004,2005,2006,2007,2008,
;; 2009,2010,2011,2012 Free Software Foundation, Inc.
;;
;; This file is part of GNU eev.
;;
;; GNU eev is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.
;;
;; GNU eev is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.
;;
;; Author:     Eduardo Ochs <eduardoochs@gmail.com>
;; Maintainer: Eduardo Ochs <eduardoochs@gmail.com>
;; Version:    2012mar05
;; Keywords:   e-scripts, help, hyperlinks, hypertext, processes,
;;             shell, tex
;;
;; Latest version: <http://angg.twu.net/eev-current/eev.el>
;;       htmlized: <http://angg.twu.net/eev-current/eev.el.html>
;;       See also: <http://angg.twu.net/eev-current/README.html>
;;            and: <http://angg.twu.net/eev-current/eev-langs.el.html>




;; `find-escratchbuffer', `find-eoutput', `find-eeffect' are obsolete -
;; use `find-eoutput-reuse' and `find-eoutput-rerun' instead.
;;
(defun find-escratchbuffer (buffer-name &rest pos-spec-list)
  "Hyperlink to an empty scratch buffer named BUFFER-NAME.
If a buffer named BUFFER-NAME exists then try to kill it with
`kill-buffer' to create a scratch buffer in its place. If the
buffer contains something precious then `kill-buffer' will ask
the user for confirmation; if the user decides not to kill the
buffer then this function aborts with an error."
  (if (get-buffer buffer-name)
      (if (kill-buffer buffer-name)
	  (switch-to-buffer buffer-name)
	(error "Not killing the buffer %s" buffer-name))
    (switch-to-buffer buffer-name)))

(defun find-eoutput (code &rest pos-spec-list)
  "Hyperlink to the output of running CODE in an empty scratch buffer.
CODE is run again every time this hyperlink if followed; compare
with `find-eeffect' and `find-sh'."
  (find-escratchbuffer (or ee-buffer-name "*output*"))
  ;; (delete-region (point-min) (point-max))
  (eval code)
  (goto-char (point-min))
  (apply 'ee-goto-position pos-spec-list))

(defun find-eeffect (ee-buffer-name code &rest pos-spec-list)
  "Hyperlink to the effect of running CODE in Emacs.
If the buffer EE-BUFFER-NAME does not exist then create it and
run CODE there; if the buffer already exists then reuse it -
suppose that its contents are already the result of running CODE
and do not run CODE again."
  (if (get-buffer ee-buffer-name)
      (apply 'find-ebuffer ee-buffer-name pos-spec-list)
    (find-ebuffer ee-buffer-name)
    (eval code)
    (set-buffer-modified-p nil)
    (goto-char (point-min))
    (apply 'ee-goto-position pos-spec-list)))




;; The two functions below bypass calling a shell and instead run
;; external programs directly. Note that the suffix "0" means "do not
;; create a new buffer, just display the result in the echo area or
;; return it", and the suffix "00" means "be even more low-level than
;; with "0": do not strip out the final newline".

;; Note: these functions are much faster than find-sh on some
;; badly-designed OSs.
;; I commented out these definitions because these functions
;; are defined in a better way below.

' (defun find-callprocess00 (prog &rest args)
    "Example: (find-callprocess00 \"lua\" \"-e\" \"print(1+2)\")"
    (with-output-to-string
      (with-current-buffer
	  standard-output
	(apply 'call-process prog nil t nil args))))

' (defun find-callprocess0 (&rest args)
    "Example: (find-callprocess0 \"lua\" \"-e\" \"print(1+2)\")"
    (replace-regexp-in-string "\n$" "" (apply 'find-callprocess00 args)))




;;;   __ _           _          _       _    __                                
;;;  / _(_)_ __   __| |      __| |_   _(_)  / / __  ___ _ __   __ _  __ _  ___ 
;;; | |_| | '_ \ / _` |____ / _` \ \ / / | / / '_ \/ __| '_ \ / _` |/ _` |/ _ \
;;; |  _| | | | | (_| |____| (_| |\ V /| |/ /| |_) \__ \ |_) | (_| | (_| |  __/
;;; |_| |_|_| |_|\__,_|     \__,_| \_/ |_/_/ | .__/|___/ .__/ \__,_|\__, |\___|
;;;                                          |_|       |_|          |___/      
;;; hyperlinks to pages in dvi/ps/pdf documents
;;;

(defun ee-dvipage (fname &optional page xdviargs)
  `("xdvi" ,@xdviargs ,@(if page (list (format "+%d" page))) ,fname))
(defun ee-pspage (fname &optional page gvargs)
  `("gv" ,@gvargs ,@(if page (list (format "--page=%d" page))) ,fname))
(defun ee-xpdfpage (fname &optional page xpdfargs)
  `("xpdf" ,@xpdfargs ,fname ,(format "%s" (or page 1))))
(defun ee-djvupage (fname &optional page)
  `("djview" ,@(if page (list (format "--page=%d" page))) ,fname))

(defun find-dvipage  (fname &optional page &rest rest)
  (find-bgprocess (ee-dvipage fname page)))
(defun find-pspage   (fname &optional page &rest rest)
  (find-bgprocess (ee-pspage fname page)))
(defun find-xpdfpage (fname &optional page &rest rest)
  (find-bgprocess (ee-xpdfpage fname page)))
(defun find-djvupage  (fname &optional page &rest rest)
  (find-bgprocess (ee-djvupage fname page)))

(defun find-pdftotext (fname &rest rest)
  (apply 'find-sh (format "pdftotext -layout %s -" fname) rest))
(defun find-pstotext (fname &rest rest)
  (apply 'find-sh (format "cat %s | pstotext" fname) rest))
(defun find-zpstotext (fname &rest rest)
  (apply 'find-sh (format "zcat %s | pstotext" fname) rest))
(defun find-djvutotext (fname &rest rest)
  (apply 'find-sh (format "djvutxt %s" fname) rest))

;;;                _                         __  _       _ 
;;;   ___ ___   __| | ___       _ __  ___   / /_| |_   _(_)
;;;  / __/ _ \ / _` |/ _ \_____| '_ \/ __| / / _` \ \ / / |
;;; | (_| (_) | (_| |  __/_____| |_) \__ \/ / (_| |\ V /| |
;;;  \___\___/ \__,_|\___|     | .__/|___/_/ \__,_| \_/ |_|
;;;                            |_|                         
;;;
;;; code-ps/dvi: mass-producing hyperlink functions

(defun ee-code-ps (code psfile)
  (format "
    (defun find-%spage (&optional n &rest comments) (interactive)
      (find-pspage %S n))
  " code psfile))
(defun ee-code-dvi (code dvifile)
  (format "
    (defun find-%spage (&optional n &rest comments) (interactive)
      (find-dvipage %S n))
  " code dvifile))
(defun ee-code-xpdf (code pdffile)
  (format "
    (defun find-%spage (&optional n &rest comments) (interactive)
      (find-xpdfpage %S n))
  " code pdffile))
(defun ee-code-djvu (code djvufile)
  (format "
    (defun find-%spage (&optional n &rest comments) (interactive)
      (find-djvupage %S n))
  " code djvufile))
(defun ee-code-pdftotext (code pdffile)
  (format "
    (defun find-%stext (&rest rest) (interactive)
      (apply 'find-pdftotext %S rest))
  " code pdffile))
(defun ee-code-pstotext (code psfile)
  (format "
    (defun find-%stext (&rest rest) (interactive)
      (apply 'find-pstotext %S rest))
  " code psfile))
(defun ee-code-zpstotext (code zpsfile)
  (format "
    (defun find-%stext (&rest rest) (interactive)
      (apply 'find-zpstotext %S rest))
  " code zpsfile))
(defun ee-code-djvutotext (code djvufile)
  (format "
    (defun find-%stext (&rest rest) (interactive)
      (apply 'find-djvutotext %S rest))
  " code djvufile))

(defun code-ps         (c f) (eval (ee-read (ee-code-ps        c f))))
(defun code-dvi        (c f) (eval (ee-read (ee-code-dvi       c f))))
(defun code-xpdf       (c f) (eval (ee-read (ee-code-xpdf      c f))))
(defun code-djvu       (c f) (eval (ee-read (ee-code-djvu      c f))))
(defun code-pdftotext  (c f) (eval (ee-read (ee-code-pdftotext c f))))
(defun code-pstotext   (c f) (eval (ee-read (ee-code-pstotext  c f))))
(defun code-zpstotext  (c f) (eval (ee-read (ee-code-zpstotext c f))))
(defun code-djvutotext (c f) (eval (ee-read (ee-code-djvutotext c f))))

(defun find-code-ps        (c f) (find-estring-elisp (ee-code-ps        c f)))
(defun find-code-dvi       (c f) (find-estring-elisp (ee-code-dvi       c f)))
(defun find-code-xpdf      (c f) (find-estring-elisp (ee-code-xpdf      c f)))
(defun find-code-djvu      (c f) (find-estring-elisp (ee-code-djvu      c f)))
(defun find-code-pdftotext (c f) (find-estring-elisp (ee-code-pdftotext c f)))
(defun find-code-pstotext  (c f) (find-estring-elisp (ee-code-pstotext  c f)))
(defun find-code-zpstotext (c f) (find-estring-elisp (ee-code-zpstotext c f)))
(defun find-code-djvutotext (c f) (find-estring-elisp (ee-code-djvutotext c f)))

;; An example of usage:
;; (brep            "http://www.tannerlectures.utah.edu/lectures/documents/Coetzee99.pdf")
;; (find-psne-links "http://www.tannerlectures.utah.edu/lectures/documents/Coetzee99.pdf")
;; (eev        "psne http://www.tannerlectures.utah.edu/lectures/documents/Coetzee99.pdf")
;; (find-pspage    "$S/http/www.tannerlectures.utah.edu/lectures/documents/Coetzee99.pdf")
;; (code-ps        "livesofanimals" "$S/http/www.tannerlectures.utah.edu/lectures/documents/Coetzee99.pdf")
;; (code-pdftotext "livesofanimals" "$S/http/www.tannerlectures.utah.edu/lectures/documents/Coetzee99.pdf")
;; (find-livesofanimalspage 1)
;; (find-livesofanimalstext)
;; (find-livesofanimalspage (+ -110 147) "Rilke's panther")
;; (find-livesofanimalstext        "147" "Rilke's panther")
;; "The jaguar's vision, unlike the panther's, is not blunted. On the
;; contrary, his eyes drill through the darkness of space. The cage
;; has no reality to him, he is elsewhere."



;; 2007nov18: I'm commenting out the block below -
;; it contains old code that is being replaced by
;; the code above.
;; The new code doesn't have docstrings yet, and I'm
;; planning to adapt some of the old docstrings for
;; the new functions - so I can't just delete this.
'(

(defun find-dvipage-old (fname &optional n &rest ignore)
  "Write into $EE a command that opens the dvi file FNAME at page N.
The command is \"xdvi +N FNAME\".
See `eev'."
  (interactive "fDVI file: ")
  (let ((command (format "xdvi +%d %s &" (or n 1) fname)))
    (eev command nil)
    command))

;; Note added in 2008mar03: according to jemarch (gv's maintainer),
;; gv 3.5.8 (and earlier versions) accept short arguments only,
;; gv 3.6.0 and gv 3.6.1 accept long arguments only,
;; gv 3.6.2 (and later versions) accept both long and short arguments.
;; To select a starting page in a "short arguments" version we have to
;; use "-page 42" instead of "--page=42".

(defun find-pspage-old (fname &optional n &rest ignore)
  "Write into $EE a command that opens the ps/pdf file FNAME at page N.
The command is \"gv -page N FNAME\".
See `eev'."
  (interactive "fPS or PDF file: ")
  (let ((command (format "gv -page %d %s &" (or n 1) fname)))
    (eev command nil)
    command))

;; (eebg-xdvi "/usr/share/doc/gdb/refcard.dvi.gz")
;; (eebg-gv "/usr/share/doc/gv/gv.ps.gz" 3 '("-scale" "-2"))
;; (eebg-channel-xterm "A" "bash" '("-gravity" "nw"))

(defun find-dvipagenow (fname &optional page xdviargs)
  "Launch an xdvi process browsing the dvi file FNAME, starting at PAGE.
The process is invoked like this: \"xdvi +PAGE $XDVIARGS FNAME\".
XDVIARGS is a list of strings - by default ().
It runs in background and its output (including error messages)
goes to the buffer \"*Messages*\"."
  (interactive "fdvi file: ")
  (apply 'start-process "xdvi" "*Messages*"
	 `("xdvi"
	   ,@(if page (list (format "+%d" page)))
	   ,@xdviargs
	   ,(ee-expand fname))))

(defun find-pspagenow (fname &optional page gvargs)
  "Launch a gv process browsing the ps or pdf file FNAME, starting at PAGE.
The process is invoked like this: \"gv --page=N $GVARGS FNAME\".
GVARGS is a list of strings - by default ().
It runs in background and its output (including error messages)
goes to the buffer \"*Messages*\"."
  (interactive "fPS or PDF file: ")
  (apply 'start-process "gv" "*Messages*"
	 `("gv"
	  ,@(if page (list (format "--page=%d" page)))
	  ,@gvargs
	  ,(ee-expand fname))))

(defalias 'find-dvipage 'find-dvipagenow)
(defalias 'find-pspage  'find-pspagenow)

(defalias 'eebg-xdvi 'find-dvipagenow)	; backward compatibility
(defalias 'eebg-gv   'find-pspagenow)	; backward compatibility

(defun code-ps (code psfile)
  "Define a function `find-CODEpage' as a hyperlink to a page in PSFILE.
See `find-pspage', `code-c-d' and `code-dvi'.
An example:\n
  (code-ps \"foo\" \"/tmp/bar.ps\")\n
runs:\n
  (defun find-foopage (&optional n &rest comments)
     (interactive)
     (find-pspage \"/tmp/bar.ps\" n))"
  (ee-eval-read-format
   "(defun find-%spage (&optional n &rest comments) (interactive)
       (find-pspage %S n))"
   code psfile))

(defun code-dvi (code dvifile)
  "Define a function `find-CODEpage' as a hyperlink to a page in DVIFILE.
See `find-dvipage', `code-c-d' and `code-ps'.
An example:\n
  (code-dvi \"foo\" \"/tmp/bar.dvi\")\n
runs:\n
  (defun find-foopage (&optional n &rest comments)
    (interactive)
    (find-dvipage \"/tmp/bar.dvi\" n))"
  (ee-eval-read-format
   "(defun find-%spage (&optional n &rest comments) (interactive)
       (find-dvipage %S n))"
   code dvifile))

)
;; End of a block of old code commented out.




(defun code-c-d-register (c d)
  "Add the pair (C D) to `code-c-d-list'.
If `code-c-d-list' already has an entry with the same C delete
the previous one. The new entry is always added to the beginning
of the list."
  (let ((c-d (assoc c code-c-d-list)))
    (if c-d (setq code-c-d-list (delq c-d code-c-d-list))))
  (setq code-c-d-list (cons (list c d) code-c-d-list)))

(defun ee-eval-read-format (formatstr &rest rest)
  "Generate a string from FORMATSTR and REST, then read and eval it.
The resulting string can contain several sexps - this function
wraps it in a \"(progn ...)\" to handle that.\n
If `ee-arg' is not nil then insert the resulting string in the
current buffer at point instead of evaluating it; see
`find-code-c-d' for an use of that."
  (let ((s (apply 'format formatstr     ; or (concat "(progn " formatstr ")")
		  rest)))
    (if ee-arg (insert "\n" s)		; for debugging
      (eval (read (concat "(progn\n" s "\n)"))))))

(defun code-c-d-base (c d)
  "See `code-c-d'."
  (ee-eval-read-format "(setq ee-%sdir \"%s\")" c d)
  (ee-eval-read-format "(setq ee-%stagsfile \"%sTAGS\")" c d)
  (ee-eval-read-format "
    (defun ee-%sfile (str)
      (concat (ee-expand ee-%sdir) str))" c c)
  (ee-eval-read-format "
    (defun ee-use-%s-tags ()
      (setq tags-file-name ee-%stagsfile))" c c)
  (ee-eval-read-format "
    (defun find-%sfile (str &rest pos-spec-list)
      (interactive (list \"\"))
      (ee-use-%s-tags)
      (apply 'find-fline (ee-%sfile str) pos-spec-list))" c c c)
  (ee-eval-read-format "
    (defun find-%stag (str &rest pos-spec-list)
      (ee-use-%s-tags) (apply 'ee-find-tag str pos-spec-list))" c c)
  (ee-eval-read-format "
    (defun find-%ssh (command &rest pos-spec-list)
      (apply 'ee-find-xxxsh ee-%sdir command pos-spec-list))" c c)
  (ee-eval-read-format "
    (defun find-%ssh0 (command)
      (funcall 'ee-find-xxxsh0 ee-%sdir command))" c c)
  (ee-eval-read-format "
    (defun find-%sw3m (furl &rest pos-spec-list)
      (apply 'find-w3m (ee-%sfile furl) pos-spec-list))" c c))

(defvar code-c-d-keywords
  '((:info
     (if b
	 (ee-eval-read-format "
           (setq ee-info-code %S)
           (setq ee-info-file %S)
           (defun find-%snode (nodename &rest pos-spec-list)
             (apply 'ee-find-xxxnode %S %S nodename pos-spec-list))"
          c   b   c c b))
     (code-c-d-rest (cddr rest)))
    (:anchor
     (ee-eval-read-format "
       (defun find-%s (file &rest rest)
         (apply 'find-anchor (ee-%sfile file) rest))" c c)
     (code-c-d-rest (cdr rest)))
    (:gdb
     (ee-eval-read-format "
       (defun eeb-%sgdb-start (&optional fname)
         (ee-use-%s-tags)
         (eeb-gdb-start ee-%sdir fname))" c c c)
     (code-c-d-rest (cdr rest)))
    (:linux
     (ee-eval-read-format "
       (defun find-%sconfvar (var &rest rest)
         (apply 'find-%sfile \"Documentation/Configure.help\"
           (concat \"\\n\" var \"\\n\") rest))" c c)
     (code-c-d-rest (cdr rest)))
    (:xdvi
     (ee-eval-read-format "
       (defun find-%sxdvi (fname &rest rest)
       (apply 'eebg-xdvi (ee-%sfile fname) rest))" c c))
    (:gv
     (ee-eval-read-format "
       (defun find-%sgv (fname &rest rest)
       (apply 'eebg-gv (ee-%sfile fname) rest))" c c))
    )
  "An alist of (KEYWORD . SEXP) pairs. See `code-c-d'.")

(defun code-c-d-rest (rest)
  "Process the \"REST\" of the arguments for a `code-c-d', besides the C and D.
This function works recursively by evaluating sexps from
`code-c-d-keywords'. See the documentation for `code-c-d'."
  (if rest
      (let* ((a (car rest))
	     (b (cadr rest))
	     (pair (assoc a code-c-d-keywords)))
	(if pair (eval (cons 'progn (cdr pair)))
	  (error "Not a code-c-d keyword: %S" a)))))

(defun code-c-d-old (c d &rest rest)
  "Define a lot of `find-xxxblah' and `ee-xxxblah' functions at once.
This function works by preparing strings and `read'ing and `eval'ing them.
Use `find-code-c-d' or `ee-code-c-d' to examine the generated
code without `eval'ing it.
For example:\n
  (find-code-c-d \"CODE\" \"/PATH/TO/\" \"INFONAME\" :anchor)\n
  (find-estring (ee-code-c-d \"CODE\" \"/PATH/TO/\" \"INFONAME\") \"sh0\")

If REST starts with a string then REST is replaced by (cons :info
REST) for convenience; the resulting value of REST is passed to
`code-c-d-rest', that processes it recursively.

When REST is empty the processing stops. When it is non-empty it
should start with a keyword that has an entry in the
`code-c-d-keywords' alist; `code-c-d-rest' executes the code
associated to that keyword, and that code - that might access the
values of C and D - takes as many arguments as it needs from the
beginning or REST, does something with that, and then invokes
`code-c-d-rest' again with what's left."
  (code-c-d-base c d)
  (unless (equal d "")
    (code-c-d-register c (ee-expand d)))
  (if rest
      (if (stringp (car rest))
	  (code-c-d-rest (cons :info rest))
	(code-c-d-rest rest)))
  c)

;; Compatibility:
(defun code-c-d-anchor (c d &optional i) (code-c-d-old c d :info i :anchor))
(defun code-c-d-gdb    (c d &optional i) (code-c-d-old c d :info i :gdb))
(defun code-c-d-linux  (c d) (code-c-d-old c d :linux))

;; For debugging:
(defun ee-code-c-d-old (&rest rest)
"Like `code-c-d', but returning the code as a string instead of evaluating it.
See `ee-eval-read-format' and `find-code-c-d'."
  (with-temp-buffer
    (let ((ee-arg 1))
      (apply 'code-c-d rest)
      (buffer-substring (point-min) (point-max)))))

(defun find-code-c-d-old (&rest rest)
"Show the code that a `code-c-d' call would evaluate, without evaluating it.
This function processes its arguments in exactly the same way as
`code-c-d' does; it doesn't accept a pos-spec-list."
  (find-estring (apply 'ee-code-c-d-old rest)))

;; 2007nov18/2007nov27:
;; If you really need the old behavior, use this:
;; (defalias 'code-c-d      'code-c-d-old)
;; (defalias 'ee-code-c-d   'ee-code-c-d-old)
;; (defalias 'find-code-c-d 'find-code-c-d-old)
;;
(defalias 'code-c-d      'code-c-d-new)
(defalias 'find-code-c-d 'find-code-c-d-new)

;; In a few places at my .emacs I was calling `code-c-d' with keywords
;; that I have not ported to `code-c-d-new' yet... for example:
;; (code-c-d     "source2e" "~/usrc/latex/base/" :xdvi)
;; At each of these calls I had to add a suffix `-old', like this:
;; (code-c-d-old "source2e" "~/usrc/latex/base/" :xdvi)



;;;                _                          _                           
;;;   ___ ___   __| | ___        ___       __| |      _ __   _____      __
;;;  / __/ _ \ / _` |/ _ \_____ / __|____ / _` |_____| '_ \ / _ \ \ /\ / /
;;; | (_| (_) | (_| |  __/_____| (_|_____| (_| |_____| | | |  __/\ V  V / 
;;;  \___\___/ \__,_|\___|      \___|     \__,_|     |_| |_|\___| \_/\_/  
;;;                                                                       
;;; 2007sep27: code-c-d (soon to be renamed "code-c-d-old") is the
;;; only part of the core of eev whose code I really hate... It will
;;; be replaced by this rewrite soon.
;;; Note that I consider replacing the old code-c-d so important that
;;; I'm adding this code here even without docstrings and without
;;; testing!...

(defun find-code-c-d-new (&rest rest)
  "Use this to inspect the code that a `code-c-d-new' would run."
  (find-estring-elisp (ee-concat (apply 'ee-code-c-d-new rest))))

(defun find-code-c-d-rest (c d &rest rest)
  "Use this to inspect the code that an `ee-code-c-d-rest' would produce."
  (find-estring-elisp (ee-concat (ee-code-c-d-rest rest))))

;; Tests:
;; (find-code-c-d-new  "lua51" "~/usrc/lua-5.1.2/")
;; (find-code-c-d-new  "lua51" "~/usrc/lua-5.1.2/" :anchor)
;; (find-code-c-d-rest "lua51" "~/usrc/lua-5.1.2/" :anchor)

(defun code-c-d-new (c d &rest rest)
  (unless (equal d "") (code-c-d-register c (ee-expand d)))
  (eval (ee-read (ee-concat (apply 'ee-code-c-d-new c d rest)))))

(defun ee-code-c-d-new (c d &rest rest)
  (if (stringp (car rest))	   	; if REST starts with a string then
      (setq rest (cons :info rest)))	; it's a shorthand for c d :info str
  (cons (ee-code-c-d-base c d)
	(ee-code-c-d-rest rest)))	; recurse over all other argments

;; Recurse over the "rest" of the arguments using `ee-code-c-d-rest'.
;; Note that the ee-code-c-d-:keyword functions can access the
;; values of c and d from the parameter list of `ee-code-c-d'.

(defun ee-code-c-d-rest (rest)
  (if rest
      (let ((fun (cdr (assq (car rest) code-c-d-keywords-new))))
	(if (not fun)
	    (error "In (code-c-d %S ...): not a keyword: %S"
		   c (car rest)))
	(apply fun (cdr rest)))))	; and fun calls ee-code-c-d-rest again

(defvar code-c-d-keywords-new nil)
(setq   code-c-d-keywords-new
 '((:info   . ee-code-c-d-:info)
   (:linfo  . ee-code-c-d-:linfo)	; new
   (:anchor . ee-code-c-d-:anchor)
   (:grep   . ee-code-c-d-:grep)
   (:wget   . ee-code-c-d-:wget)
   ))

(defun ee-code-c-d-base (c d)
  (format "
   (setq ee-%sdir \"%s\")
   (setq ee-%stagsfile \"%sTAGS\")
   (defun ee-%sfile (str)
     (concat (ee-expand ee-%sdir) str))
   (defun ee-use-%s-tags ()
     (setq tags-file-name ee-%stagsfile))
   (defun find-%sfile (str &rest pos-spec-list)
     (interactive (list \"\"))
     (ee-use-%s-tags)
     (apply 'find-fline (ee-%sfile str) pos-spec-list))
   (defun find-%stag (str &rest pos-spec-list)
     (ee-use-%s-tags)
     (apply 'ee-find-tag str pos-spec-list))
   (defun find-%ssh (command &rest pos-spec-list)
     (apply 'ee-find-xxxsh ee-%sdir command pos-spec-list))
   (defun find-%ssh0 (command)
     (funcall 'ee-find-xxxsh0 ee-%sdir command))
   (defun find-%ssh00 (command)
     (funcall 'ee-find-xxxsh00 ee-%sdir command))
   (defun find-%sw3m (furl &rest pos-spec-list)
     (apply 'find-w3m (ee-%sfile furl) pos-spec-list))
   (defun find-%sgrep (grep-command-args &rest pos-spec-list)
     (apply 'ee-find-grep ee-%sdir grep-command-args pos-spec-list))
   " c d				; (setq ee-%sdir ...)
     c d				; (setq ee-%stagsfile ...)
     c c				; (defun ee-%sfile ...)
     c c				; (defun ee-use-%s-tags ...)
     c c c				; (defun find-%sfile ...)
     c c				; (defun find-%stag ...)
     c c				; (defun find-%ssh ...)
     c c				; (defun find-%ssh0 ...)
     c c				; (defun find-%ssh00 ...)
     c c				; (defun find-%sw3m ...)
     c c))				; (defun find-%sgrep ...)

(defun ee-code-c-d-:info (info &rest rest)            ; from mini
 (cons (format "
   (defun find-%snode (page &rest pos-spec-list)
     (interactive (list \"\"))
     (setq ee-info-code %S)
     (setq ee-info-file %S)    ;; a (temporary) hack for M-h M-i
     (apply 'find-node (format \"(%s)%%s\" page) pos-spec-list))
   " c c info info) (ee-code-c-d-rest rest)))

(defun ee-code-c-d-:linfo (manual &rest rest)
 (cons (format "
   (defun find-%snode (section &rest pos-spec-list)
     (interactive (list \"\"))
     (apply 'ee-find-node ee-%sdir %S section pos-spec-list))
   " c c manual) (ee-code-c-d-rest rest)))

(defun ee-code-c-d-:anchor (&rest rest)
 (cons (format "
   (defun find-%s (str &rest pos-spec-list)
     (apply 'find-anchor (ee-%sfile str) pos-spec-list))
   " c c) (ee-code-c-d-rest rest)))

(defun ee-code-c-d-:grep (&rest rest)
 (cons (format "
   ;; (defun find-%sgrep (grep-command-args &rest pos-spec-list)
   ;;   (apply 'ee-find-grep ee-%sdir grep-command-args pos-spec-list))
   " c c) (ee-code-c-d-rest rest)))

(defun ee-code-c-d-:wget (url &rest rest)
 (cons (format "
   (defun ee-%surl (semiurl) (concat %S semiurl))
   (defun find-%swget (semiurl &rest pos-spec-list)
     (interactive (list \"\"))
     (apply 'find-wget (ee-%surl semiurl) pos-spec-list))
   " c url c c) (ee-code-c-d-rest rest)))

;; We can define new keywords for code-c-d by adding a pair entries to 
;; code-c-d-keywords and An example of extension...
;;
;; (add-to-alist 'code-c-d-keywords '(:linfo  . ee-code-c-d-:linfo))


;;;           _           _                 _          _                      
;;; __      _(_)_ __   __| | _____      __ | |__   ___| |__   __ ___   ___ __ 
;;; \ \ /\ / / | '_ \ / _` |/ _ \ \ /\ / / | '_ \ / _ \ '_ \ / _` \ \ / / '__|
;;;  \ V  V /| | | | | (_| | (_) \ V  V /  | |_) |  __/ | | | (_| |\ V /| |   
;;;   \_/\_/ |_|_| |_|\__,_|\___/ \_/\_/   |_.__/ \___|_| |_|\__,_| \_/ |_|   
;;;                                                                           
;;; setting `pop-up-windows' to nil inside eev-mode
;;; (what about setting debug-on-error and friends to nil too?)
;;; See: (find-elnode "Displaying Buffers" "pop-to-buffer")
;;;      (find-elnode "Choosing Window" "pop-up-windows")
;;;

(defun ee-setq-to-current-1 (var)
  "If the current value of VAR is VALUE then return (setq VAR 'VALUE).
Example: (let ((a 22)) (ee-setq-to-current-1 'a))"
  `(setq ,var ',(symbol-value var)))

(defun ee-setq-to-current (&rest vars)
  "Example: (find-epp (let ((a 22) (b 33)) (ee-setq-to-current 'a 'b)))
This function is used by `eev-mode' (actually from the code in
the variable `eev-mode-global-settings-saver') to create a sexp
that when eval'ed will restore the current value of some global
variable(s)."
  (cons 'progn (mapcar 'ee-setq-to-current-1 vars)))

;; Ideas for names:
;; eev-vars-setup
;; eev-vars-setup-code
;; eev-vars-reset-code

(defvar eev-mode-global-settings-restorer nil)
(defvar eev-mode-global-settings-saver
  '(progn (setq eev-mode-global-settings-restorer
		(ee-setq-to-current 'pop-up-windows))
	  (setq pop-up-windows nil))
  "If you don't want eev-mode to set any global flags on entry then set this to nil.
The default value for this variable is some Lisp code that sets
the global value of `pop-up-windows' to nil when `eev-mode' is
entered and restores the old value when the mode is left.

Several functions in Emacs can be used as elisp hyperlinks, but
some of them try to follow one convention on splitting windows,
while other follow another one. Let's call those conventions the
\"display help buffer\" convention and the \"follow hyperlink\"
convention; the \"display help buffer\" convention is to keep the
current buffer visible and at the same time display the target of
the hyperlink in another window, splitting the current window if
needed; the \"follow hyperlink\" convention is to never split
windows or switch a different window, and always replace the
buffer in the current window by the target of the hyperlink.

The \"follow hyperlink\" convention makes it easier to return
from hyperlinks: after following a hyperlink just run
`ee-kill-this-buffer' (bound to
`\\<eev-mode-map>\\[ee-kill-this-buffer]' in eev-mode) or
`bury-buffer' (`\\<eev-mode-map>\\[bury-buffer]' in eev-mode).

All hyperlink functions defined by eev follow the \"follow
hyperlink\" convention, and if we set `pop-up-windows' to nil
then most standard Emacs functions usable as hyperlinks also
behave in the \"follow hyperlink\" way.

If you don't want eev-mode to change global variables then do
this (with eev-mode off):

    (setq eev-mode-global-settings-saver nil)")

(put 'eev-mode-global-settings-saver    'risky-local-variable t)
(put 'eev-mode-global-settings-restorer 'risky-local-variable t)

(defun eev-mode-global-settings-set ()
  "This is run by `eev-mode' when eev-mode is turned on.
See the documentation of `eev-mode-global-settings-saver'."
  (if eev-mode-global-settings-restorer
      (message "Not overwriting `eev-mode-global-settings-restorer'.")
    (eval eev-mode-global-settings-saver)))

(defun eev-mode-global-settings-restore ()
  "This is run by `eev-mode' when eev-mode is turned off.
See the documentation of `eev-mode-global-settings-saver'."
  (eval eev-mode-global-settings-restorer)
  (setq eev-mode-global-settings-restorer nil))




'(

;; Old code, that will be deleted soon.

;; Some people have told me that their window managers binds `M-h'...
;; How do I create a secondary key that acts like the `M-h' prefix?

(defvar eev-mode-map nil)
(if eev-mode-map
    ()
(setq eev-mode-map (make-sparse-keymap))
(define-key eev-mode-map "\M-E" 'eek-eval-last-sexp)    ; extends     C-x C-e
(define-key eev-mode-map "\M-e" 'eek-eval-sexp-eol)     ; extends C-e C-x C-e
(define-key eev-mode-map "\M-k" 'kill-this-buffer)      ; convenience
(define-key eev-mode-map "\M-K" 'bury-buffer)           ; convenience
(define-key eev-mode-map [f3]   'eeb-default)
;; (define-key eev-mode-map [f8] 'eewalk-do-walk-or-set) ; in eev-walk.el
(define-key eev-mode-map [f8]   'eepitch-this-line)
(define-key eev-mode-map [f9]   'eechannel-do-this-line)
(define-key eev-mode-map [f12]  'eesteps-do-step)
(define-key eev-mode-map "\M-P" 'ee-yank-one-line)
(define-key eev-mode-map "\M-?" 'eev-help-page)
(define-key eev-mode-map "\M-G" 'eegud-show-gud-buffer)	; hack, in eev-insert.el

(define-key eev-mode-map "\M-h\M-d" 'find-debpkg-links)    ; in eev-insert.el
(define-key eev-mode-map "\M-h\M-f" 'find-efunction-links) ; in eev-insert.el
(define-key eev-mode-map "\M-h\M-i" 'find-einfo-links)     ; in eev-insert.el
(define-key eev-mode-map "\M-h\M-k" 'find-ekey-links)      ; in eev-insert.el
(define-key eev-mode-map "\M-h\M-m" 'find-manpage-links)   ; in eev-insert.el
(define-key eev-mode-map "\M-h\M-v" 'find-evariable-links) ; in eev-insert.el
(define-key eev-mode-map "\M-hf"    'find-file-links)      ; in eev-insert.el
(define-key eev-mode-map "\M-hm"    'find-last-manpage-links) ;in eev-insert.el
(define-key eev-mode-map "\M-h\M-y" 'eemklinks-yank-pos-spec) ;in eev-insert.el
(define-key eev-mode-map "\M-h2"    'eemklinks-duplicate-this-line) ; idem
(define-key eev-mode-map "\M-h\M-2" 'eemklinks-duplicate-this-line) ; idem
(define-key eev-mode-map "\M-I"     'ee-ill)               ; in eev-insert.el

(define-key eev-mode-map "\M-h\M-c" 'describe-char)
(define-key eev-mode-map "\M-h\M-t" 'find-etpat)
(define-key eev-mode-map "\M-ht"    'find-etpat0)
(define-key eev-mode-map "\M-h\M-s" 'find-efacedescr)

(define-key eev-mode-map [?\C-,] 'eev-compose-two-keys)     ; only works on X
(define-key eev-mode-map [?\M-,] 'eev-compose-two-keys)
)

)






;; Local Variables:
;; coding:            raw-text-unix
;; ee-anchor-format:  "%s"
;; ee-anchor-format:  "defun %s "
;; ee-comment-prefix: ";;"
;; no-byte-compile:   t
;; End:
